from queue import Queue
from sys import stdout

from aiohttp import ClientSession

from air_kit.process import Process

telemetry = None


def get_telemetry():
    global telemetry
    return telemetry


def set_telemetry(telemetry_new):
    global telemetry
    telemetry = telemetry_new


class TelemetryStream:
    def __init__(
        self,
        is_stdout: bool = True,
        is_telemetry: bool = True,
    ):
        self.is_stdout = is_stdout
        self.is_telemetry = is_telemetry

    def write(self, data):
        if self.is_stdout:
            stdout.write(data)

        if self.is_telemetry and get_telemetry():
            get_telemetry().enqueue(data)

    def flush(self):
        if self.is_stdout:
            stdout.flush()


class TelemetryProcess(Process):
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.queue = Queue()

        super().__init__("telemetry", period=3.0)

    def enqueue(self, data):
        self.queue.put(data)

    async def _on_start(self) -> None:
        pass

    async def _on_stop(self) -> None:
        await self._flush()

    async def _activity(self):
        await self._flush()

    async def _flush(self):
        if self.queue.empty():
            return

        buffer = []

        while not self.queue.empty():
            buffer.append(self.queue.get()[:-1])

        # TODO: check response ok
        # TODO: handle exceptions

        async with ClientSession() as session:
            async with session.post(
                f"http://{self.host}:{self.port}/logs",
                json={
                    "items": buffer,
                },
            ) as _:
                pass



from logging import getLogger
from time import time_ns

logger = getLogger(__name__)


class Profiler:
    def __init__(self, message, arguments=None, *, enabled=True):
        if arguments is None:
            arguments = {}

        self.message = message
        self.arguments = arguments
        self.enabled = enabled
        self.time_start = None

    def __enter__(self):
        if self.enabled:
            self.time_start = time_ns()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.enabled:
            self.arguments["time_delta"] = (
                time_ns() - self.time_start
            ) / 1000000000
            logger.info(self.message.format(**self.arguments))

        return exc_type is None



from asyncio import gather, get_event_loop, sleep
from asyncio.runners import _cancel_all_tasks
from dataclasses import dataclass
from logging import getLogger
from os import environ
from signal import SIGHUP, SIGINT, SIGTERM, SIGUSR1, SIGUSR2
from time import time

from dotenv import load_dotenv
from yaml import dump, load

from .environment import Environment
from .errors import ConfigurationValidationError

logger = getLogger(__name__)


def read_yaml(directory, file):
    with open(f"{directory}/{file}", "rt") as handle:
        return load(handle.read())


def write_yaml(directory, file, data):
    with open(f"{directory}/{file}", "wt") as handle:
        handle.write(dump(data, indent=2))


@dataclass
class EnvironmentVariable:
    key: str
    cast: type
    default: object
    is_required: bool


class BaseConfiguration:
    def __init__(self):
        self.environment = environ.get(
            "ENVIRONMENT",
            Environment.DEVELOPMENT,
        ).lower()
        self.configuration = self._merge_configuration(
            self._load_features(),
            self._load_options(),
            self._load_environment(),
        )
        if not self._validate_configuration(self.configuration):
            raise ConfigurationValidationError()

    def _load_features(self):
        # fmt: off
        return {
            name: read_yaml("configuration/features", value)
            for name, value in self._feature_files().items()
        }
        # fmt: on

    def _load_options(self):
        # fmt: off
        return {
            name: read_yaml("configuration/options", value)
            for name, value in self._option_files().items()
        }
        # fmt: on

    def _load_environment(self):
        # fmt: off
        load_dotenv("configuration/environments/.env_{}".format(self.environment))
        variables = {}

        for environment_variable in self._environment_variables():
            key_upper = environment_variable.key.upper()

            if key_upper in environ:
                value = environment_variable.cast(environ[key_upper])
            else:
                if environment_variable.is_required:
                    raise Exception("Value is required.")  # TODO: error type

                value = environment_variable.default

            variables[environment_variable.key] = value

        return variables
        # fmt: on

    def _feature_files(self):
        return {}

    def _option_files(self):
        return {}

    def _environment_variables(self):
        return ()

    def _merge_configuration(self, features, options, environment):
        return {}

    def _validate_configuration(self, configuration) -> bool:
        return True


class Runner:
    def __init__(self, applications):
        self.applications = applications

        for application in self.applications:
            application.runner = self

    def start(self):
        event_loop = get_event_loop()

        try:
            return event_loop.run_until_complete(self.run())
        except BaseException:
            logger.exception("Exception in runner loop.")
            pass

        try:
            _cancel_all_tasks(event_loop)
            event_loop.run_until_complete(event_loop.shutdown_asyncgens())
        except BaseException:
            logger.exception("Exception in runner shutdown.")
            pass

    def stop(self):
        for application in self.applications:
            application.stop()

    async def run(self):
        await gather(*[application.run() for application in self.applications])


class BaseApplication:
    def __init__(self, name, *, period=1.0):
        self.name = name
        self.period = period
        self.tasks = []
        self.runner = None
        self.is_active = True

    def start(self):
        event_loop = get_event_loop()

        try:
            event_loop.run_until_complete(self.run())
        except BaseException:
            logger.exception(f"Exception in application {self.name} loop.")
            pass

        try:
            _cancel_all_tasks(event_loop)
            event_loop.run_until_complete(event_loop.shutdown_asyncgens())
        except BaseException:
            logger.exception(f"Exception in application {self.name} shutdown.")
            pass

    def stop(self):
        self.is_active = False

    async def run(self):
        # Signal handlers.
        def _on_exit_signal():
            if self.runner is None:
                self.stop()
                return

            self.runner.stop()

        event_loop = get_event_loop()
        event_loop.add_signal_handler(SIGINT, _on_exit_signal)
        event_loop.add_signal_handler(SIGTERM, _on_exit_signal)
        event_loop.add_signal_handler(SIGUSR1, self._on_usr_1)
        event_loop.add_signal_handler(SIGUSR2, self._on_usr_2)
        event_loop.add_signal_handler(SIGHUP, self._on_hup)

        # Start.
        await self._on_start()
        logger.debug(f"Application {self.name} started.")

        # Workers.
        async def _worker(worker):
            while True:
                await sleep(worker.period - time() % worker.period)

                event_loop = get_event_loop()
                event_loop.create_task(worker.function())

        for worker in self._workers():
            task = event_loop.create_task(_worker(worker))
            self.tasks.append(task)

        # Loop.
        while self.is_active:
            try:
                await self._activity()
                await sleep(self.period)
            except BaseException:
                logger.exception(f"Exception while running application {self.name}.")
                await sleep(1)

        # Stop.
        await self._on_stop()
        logger.debug(f"Application {self.name} stopped.")

    async def _on_start(self):
        pass

    async def _on_stop(self):
        pass

    async def _activity(self):
        pass

    def _on_usr_1(self):
        pass

    def _on_usr_2(self):
        pass

    def _on_hup(self):
        pass

    def _workers(self):
        return []



from logging import getLogger

logger = getLogger(__name__)


def exclusive(function: callable):
    async def _wrapper(self, *args, **kwargs):
        async with self.lock:
            return await function(*args, **kwargs)

    return _wrapper



from asyncio import run, sleep
from logging import getLogger
from multiprocessing import Process as BaseProcess
from multiprocessing import Value

logger = getLogger(__name__)


class Process(BaseProcess):
    def __init__(self, name: str, *, period: float = 1.0, **kwargs):
        self.name = name
        self.period = period
        self.is_active = Value("b", 1)

        super().__init__(name=name, args=(self.is_active,), kwargs=kwargs)

    def start(self) -> None:
        super().start()

    def stop(self) -> None:
        self.is_active.value = 0
        self.join()

    def run(self) -> None:
        try:
            run(self._run())
        except BaseException:
            logger.exception(f"Exception in process {self.name} loop.")
            pass

    async def _run(self) -> None:
        # Getting multiprocess parameters.
        self.is_active = self._args[0]

        for name, value in self._kwargs.items():
            setattr(self, name, value)

        # Start.
        await self._on_start()
        logger.debug(f"Process id {self.pid}, name {self.name} started.")

        # Running.
        await self._loop()

        # Stop.
        await self._on_stop()
        logger.debug(f"Process id {self.pid}, name {self.name} stopped.")

    async def _loop(self) -> None:
        while self.is_active.value:
            try:
                await self._activity()
                await sleep(self.period)
            except BaseException:
                logger.exception(
                    f"Exception while running process id {self.pid}, name {self.name}."
                )
                await sleep(1)

    async def _on_start(self) -> None:
        pass

    async def _on_stop(self) -> None:
        pass

    async def _activity(self) -> None:
        pass



from contextvars import ContextVar
from datetime import datetime
from json import dumps
from logging import DEBUG, INFO, Filter, Formatter
from logging.config import dictConfig
from sys import stdout

correlation_id = ContextVar("correlation_id", default="")


def get_correlation_id():
    global correlation_id
    return correlation_id.get()


def set_correlation_id(correlation_id_new):
    global correlation_id
    correlation_id.set(correlation_id_new)


class CustomFilter(Filter):
    def filter(self, record):
        return True


class BaseFormatter(Formatter):
    def formatTime(self, record, datefmt=None):
        created = datetime.utcfromtimestamp(record.created)
        return created.isoformat()


class JsonFormatter(BaseFormatter):
    def __init__(self, _, *args, **kwargs):
        super().__init__(
            dumps(
                {
                    "correlation_id": "%(correlation_id)s",
                    "time": "%(asctime)s",
                    "level": "%(levelname)s",
                    "path": "%(pathname)s",
                    "line": "%(lineno)s",
                    "message": "%(message)s",
                }
            ),
            *args,
            **kwargs,
        )

    def format(self, record):
        record.correlation_id = get_correlation_id()
        return super().format(record)


class TextFormatter(BaseFormatter):
    def __init__(self, _, *args, **kwargs):
        super().__init__(
            "%(correlation_id)s %(asctime)s %(levelname)s %(pathname)s %(line)s %(message)s",
            *args,
            **kwargs,
        )

    def format(self, record):
        record.correlation_id = get_correlation_id()
        return super().format(record)


def configure_loggers(
    stream: object,
    formatter: str = "json",
):
    dictConfig(
        {
            "version": 1,
            "filters": {
                "custom": {
                    "()": "air_kit.logging.CustomFilter",
                },
            },
            "formatters": {
                "json": {
                    "class": "air_kit.logging.JsonFormatter",
                },
                "text": {
                    "class": "air_kit.logging.TextFormatter",
                },
            },
            "handlers": {
                "output": {
                    "class": "logging.StreamHandler",
                    "level": DEBUG,
                    "filters": ("custom",),
                    "formatter": formatter,
                    "stream": stream,
                },
            },
            "loggers": {
                "": {
                    "level": DEBUG,
                    "filters": ("custom",),
                    "handlers": ("output",),
                    "propagate": True,
                },
                "urllib3.connectionpool": {
                    "level": INFO,
                    "filters": ("custom",),
                    "handlers": ("output",),
                    "propagate": True,
                },
                "docker.utils.config": {
                    "level": INFO,
                    "filters": ("custom",),
                    "handlers": ("output",),
                    "propagate": True,
                },
                "asyncpgsa.query": {
                    "level": INFO,
                    "filters": ("custom",),
                    "handlers": ("output",),
                    "propagate": True,
                },
            },
        }
    )



from abc import ABC, abstractmethod
from queue import Queue

from aiohttp.web import Application, AppRunner, Request, Response, TCPSite, post
from telegram import Bot, Update
from telegram.ext import Dispatcher, Updater


class Mode:
    WEBHOOK = "webhook"
    POOLING = "pooling"

class TelegramBot(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass


class PoolingTelegramBot(TelegramBot):
    def __init__(self, token: str):
        super().__init__()
        self.updater = Updater(token, use_context=True)

    async def start(self):
        self.updater.start_polling()

    async def stop(self):
        self.updater.stop()

    @property
    def dispatcher(self):
        return self.updater.dispatcher


class WebhookTelegramBot(TelegramBot):
    def __init__(self, host: str, port: int, fqdn: str, token: str):
        super().__init__()

        class Stub:
            def __init__(self):
                self.defaults = None

        bot = Bot(token)
        dispatcher = Dispatcher(Stub(), workers=0, update_queue=Queue())

        application = Application()
        application.add_routes(
            [
                post(
                    "/{token}",
                    self._index,
                ),
            ]
        )

        async def on_start(_):
            bot.setWebhook(f"https://{fqdn}/{token}")

        async def on_stop(_):
            bot.deleteWebhook()

        application.on_startup.append(on_start)
        application.on_shutdown.append(on_stop)

        self.host = host
        self.port = port

        self.bot = bot
        self.dispatcher = dispatcher

        self.application = application
        self.runner = None
        self.site = None

    async def start(self):
        # Create runner.
        self.runner = AppRunner(self.application)
        await self.runner.setup()

        # Create tcp site.
        self.site = TCPSite(self.runner, self.host, self.port)
        await self.site.start()

    async def stop(self):
        pass

    async def _index(self, request: Request) -> Response:
        self.dispatcher.process_update(
            Update.de_json(
                await request.json(),
                self.bot,
            ),
        )
        return Response()



from setuptools import setup

with open("README.md", "r") as handle:
    long_description = handle.read()

# fmt: off
configuration = dict(
    name="air_kit",
    version="0.1.4",
    description="air's kit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
    ],
    author="air64",
    author_email="contact.air64@yandex.ru",
    maintainer=", ".join(
        (
            "air64 <contact.air64@yandex.ru>",
        ),
    ),
    maintainer_email="contact.air64@yandex.ru",
    url="https://gitlab.com/air64/air_kit",
    project_urls={
        "Gitlab": "https://gitlab.com/air64/air_kit",
    },
    license="MIT",
    packages=["air_kit"],
    python_requires=">=3.7",
    install_requires=[
        "python-dotenv==0.13.0",
        "pyyaml==5.3.1",
    ],
    extras_require={
        "telegram_bot": [
            "python-telegram-bot~=13.7",
        ],
    },
    include_package_data=True,
)
setup(**configuration)
# fmt: on
